﻿------------------------------------------------------------------------------
--	FILE:	 MeteorSurvival.lua
--	AUTHOR:  djkrose
--	PURPOSE: A meteor crater in the middle of an ice planet with just a small strip of habitable land for all civilizations to batte for.
------------------------------------------------------------------------------
--	Based on: InlandSea.lua
--  Copyright (c) 2014 Firaxis Games, Inc. All rights reserved.
------------------------------------------------------------------------------

include "MapEnums"
include "MapUtilities"
include "MountainsCliffs"
include "RiversLakes"
include "FeatureGenerator"
include "TerrainGenerator"
include "NaturalWonderGenerator"
include "ResourceGenerator"
include "AssignStartingPlots"

local g_iW, g_iH;
local g_iFlags = {};
local g_continentsFrac = nil;
local g_randomWiggle = nil;
local g_count_icy_plots = 0;
local g_craterCenterX, g_craterCenterY;

-------------------------------------------------------------------------------
function GenerateMap()
	print("Generating Meteor Survival Map");
	local pPlot;

	-- Set globals
	g_iW, g_iH = Map.GetGridSize();
	g_iFlags = TerrainBuilder.GetFractalFlags();
	
	local shift = 0.07;

	-- Generate basic height map with hills, mountains, but no ocean yet
	plotTypes = GeneratePlotTypes();

	-- Hit the world with a meteor! Also adds ocean tiles to plotTypes for the crater
	terrainTypes = GenerateTerrainTypes(plotTypes);

	AddCoast(plotTypes, terrainTypes);

    -- Pass terrainTypes to TerrainBuilder
	print("Passing terrain to TerrainBuilder ...");
	for i = 0, (g_iW * g_iH) - 1, 1 do
		pPlot = Map.GetPlotByIndex(i);
		TerrainBuilder.SetTerrainType(pPlot, terrainTypes[i]);
	end
	
	-- Temp
	AreaBuilder.Recalculate();
	local biggest_area = Areas.FindBiggestArea(false);
	print("After adding hills the biggest area is " .. biggest_area:GetPlotCount() .. " of " .. g_iW * g_iH .. " plots.");

	-- River generation is affected by plot types, originating from highlands and preferring to traverse lowlands.
	AddRivers();
	
	-- Lakes would interfere with rivers, causing them to stop and not reach the ocean, if placed any sooner.
	local numLargeLakes = GameInfo.Maps[Map.GetMapSize()].Continents;
	AddLakes(numLargeLakes);

	AddFeatures();
	
	print("Adding cliffs ...");
	AddCliffs(plotTypes, terrainTypes);

	local args = {
		numberToPlace = GameInfo.Maps[Map.GetMapSize()].NumNaturalWonders,
		Invalid= {g_FEATURE_GALAPAGOS, g_FEATURE_PIOPIOTAHI},
	};
	local nwGen = NaturalWonderGenerator.Create(args);

	AreaBuilder.Recalculate();
	TerrainBuilder.AnalyzeChokepoints();
	TerrainBuilder.StampContinents();
	
	--for i = 0, (g_iW * g_iH) - 1, 1 do
		--pPlot = Map.GetPlotByIndex(i);
		--print ("i: plotType, terrainType, featureType: " .. tostring(i) .. ": " .. tostring(plotTypes[i]) .. ", " .. tostring(terrainTypes[i]) .. ", " .. tostring(pPlot:GetFeatureType(i)));
	--end
	local resourcesConfig = MapConfiguration.GetValue("resources");
	local args = {
		resources = resourcesConfig,
	};
	local resGen = ResourceGenerator.Create(args);

	print("Creating start plot database ...");
	-- START_MIN_Y and START_MAX_Y is the percent of the map ignored for major civs' starting positions.
	local startConfig = MapConfiguration.GetValue("start");-- Get the start config
	local args = {
		MIN_MAJOR_CIV_FERTILITY = 200,
		MIN_MINOR_CIV_FERTILITY = 50, 
		MIN_BARBARIAN_FERTILITY = 1,
		START_MIN_Y = 15,
		START_MAX_Y = 15,
		START_CONFIG = startConfig,
	};
	local start_plot_database = AssignStartingPlots.Create(args)

	local GoodyGen = AddGoodies(g_iW, g_iH);
end

-------------------------------------------------------------------------------
-- Input a Hash; Export width, height, and wrapX
function GetMapInitData(MapSize)
	local MapSizeTypes = {};
	local Width = 0;
	local Height = 0;

	for row in GameInfo.Maps() do
		if(MapSize == row.Hash) then
			Width = row.GridWidth;
			Height = row.GridHeight;
		end
	end

	local WrapX = false;

	return {Width = Width, Height = Height, WrapX = WrapX,}
end

-------------------------------------------------------------------------------
function GeneratePlotTypes()
	print("Generating plot types ...");
	local plotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);

	local world_age_old = 2;
	local world_age_normal = 3;
	local world_age_new = 5;
	
	AreaBuilder.Recalculate();
	
	-- Land and water are set. Now apply hills and mountains.
	
	--	local world_age
	local world_age = MapConfiguration.GetValue("world_age");
	if (world_age == 1) then
		world_age = world_age_new;
	elseif (world_age == 2) then
		world_age = world_age_normal;
	elseif (world_age == 3) then
		world_age = world_age_old;
	else
		world_age = 2 + TerrainBuilder.GetRandomNumber(4, "Random World Age - Lua");
	end

	local args = {};
	args.world_age = world_age;
	args.iW = g_iW;
	args.iH = g_iH
	args.iFlags = g_iFlags;
	args.blendRidge = 10;
	args.blendFract = 1;
	args.extra_mountains = 7 - world_age;
	mountainRatio = 13 + world_age;
	plotTypes = ApplyTectonics(args, plotTypes);
	plotTypes = AddLonelyMountains(plotTypes, mountainRatio);

	-- Plot Type generation completed. Return global plot array.

	return plotTypes;
end

-------------------------------------------------------------------------------
function GenerateTerrainTypes(plotTypes)
	local MAX_CRATER_SHIFT = 10; -- in percent of full width and height

	local terrainTypes = {};

	print("Generating terrain types for Meteor Survival map ...");

	-- Define middle of crater
	g_craterCenterX = math.floor(g_iW / 2 + g_iW * (TerrainBuilder.GetRandomNumber(MAX_CRATER_SHIFT * 2, "Creater Shift X - Lua") - MAX_CRATER_SHIFT) / 100);
	g_craterCenterY = math.floor(g_iH / 2 + g_iH * (TerrainBuilder.GetRandomNumber(MAX_CRATER_SHIFT * 2, "Creater Shift Y - Lua") - MAX_CRATER_SHIFT) / 100);
	print("Crater center: ", g_craterCenterX, g_craterCenterY);

	-- Define all terrain depending on distance to meteor crater
	for iX = 0, g_iW - 1 do
		for iY = 0, g_iH - 1 do
			local i = (iY * g_iW) + iX;

			local hilltype  = GetHillType(plotTypes[i]);
			local influence = GetCraterInfluence(iX, iY);

			if influence < 13 then
				plotTypes[i] = g_PLOT_TYPE_OCEAN;
				terrainTypes[i] = g_TERRAIN_TYPE_OCEAN;
			elseif influence < 15 then
				plotTypes[i] = g_PLOT_TYPE_OCEAN;
				terrainTypes[i] = g_TERRAIN_TYPE_COAST;
			elseif influence < 18 then
				terrainTypes[i] = g_TERRAIN_TYPE_DESERT + hilltype;
			elseif influence < 21 then
				terrainTypes[i] = g_TERRAIN_TYPE_PLAINS + hilltype;
			elseif influence < 24 then
				terrainTypes[i] = g_TERRAIN_TYPE_GRASS + hilltype;
			elseif influence < 30 then 
				terrainTypes[i] = g_TERRAIN_TYPE_TUNDRA + hilltype;
			else
				terrainTypes[i] = g_TERRAIN_TYPE_SNOW + hilltype;
				g_count_icy_plots = g_count_icy_plots + 1;
			end
		end
	end

	return terrainTypes;
end

-------------------------------------------------------------------------------
function AddCoast(plotTypes, terrainTypes)
	for iX = 0, g_iW - 1 do
		for iY = 0, g_iH - 1 do
			local index = (iY * g_iW) + iX;
			if (plotTypes[index] == g_PLOT_TYPE_OCEAN) then
				if (IsAdjacentToLand(plotTypes, iX, iY)) then
					terrainTypes[index] = g_TERRAIN_TYPE_COAST;
				end
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Returns the influence of the crater on the given x,y coordinates in percent.
-- The SMALLER the value the LARGER is the influence, with 0 being the middle of the crater.
function GetCraterInfluence(x, y)
	local WIGGLE_STRENGTH = 7; -- 0 = no wiggle; 10 = displace terrain by max 10 percent of map width
	local WIGGLE_SMOOTHNESS = 720; -- Minimum: 360 = Not smooth at all; 3600 = very smooth; higher = even smoother

	local deltaX = x - g_craterCenterX;
	local deltaY = y - g_craterCenterY;

	local distance = math.sqrt((g_craterCenterX - x) * (g_craterCenterX - x) + (g_craterCenterY - y) * (g_craterCenterY - y));

	local distancePercent = distance / g_iW * 100;

	local degree = math.atan2(deltaY, deltaX) * 180 / math.pi;

	local wiggle = WiggleFunction(degree, WIGGLE_STRENGTH, WIGGLE_SMOOTHNESS);
	--local wiggle = 0;

	local influence = distancePercent + wiggle;

	return influence;
end

-------------------------------------------------------------------------------
-- Returns 1=hill, 2=mountain, 0=neither
function GetHillType(plotType)
	local hilltype;
	if (plotType == g_PLOT_TYPE_HILLS) then
		hilltype = g_TERRAIN_BASE_TO_HILLS_DELTA;
	elseif (plotType == g_PLOT_TYPE_MOUNTAIN) then
		hilltype = g_TERRAIN_BASE_TO_MOUNTAIN_DELTA;
	else
		hilltype = 0;
	end
	return hilltype;
end

-------------------------------------------------------------------------------
-- Returns a pseudo-random wiggle value based on the given input X.
-- Return value will be between -maxHeight and +maxHeight and be stretched over maxWidth values before it repeats itself.
-- See: https://www.desmos.com/calculator/7xzsuq3xac
function WiggleFunction(x, maxHeight, maxWidth)
	if (g_randomWiggle == nil) then
		g_randomWiggle = TerrainBuilder.GetRandomNumber(2 * math.pi * 10000, "Random Wiggle - Lua") / 10000;
		print("Random wiggle factor: ", g_randomWiggle);
	end

	local q = math.sin(41 * x * 2 * math.pi / maxWidth -     41 * g_randomWiggle);
	local a = math.sin( 7 * x * 2 * math.pi / maxWidth + q +  7 * g_randomWiggle);
	local b = math.sin(11 * x * 2 * math.pi / maxWidth + 1 - 11 * g_randomWiggle);
	local c = math.sin(13 * x * 2 * math.pi / maxWidth + 2 + 13 * g_randomWiggle);
	local y = a * b * c * maxHeight;
	return y;
end

-----------------------------------------------------------------------------
function GetRiverValueAtPlot(plot)
	if(plot:IsNWOfCliff() or plot:IsWOfCliff() or plot:IsNEOfCliff()) then
		return -1;
	elseif(plot:IsNaturalWonder() or AdjacentToNaturalWonder(plot)) then
		return -1;
	end

	local x = plot:GetX()
	local y = plot:GetY()
	local random_factor = TerrainBuilder.GetRandomNumber(3, "River Rand");
	local direction_influence_value = (math.abs(x - (g_iW / 2)) + math.abs(y - (g_iH / 2))) * random_factor;
	local sum = GetPlotElevation(plot) * 20 + direction_influence_value;

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		local adjacentPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), direction);
		if (adjacentPlot) then
			sum = sum + GetPlotElevation(adjacentPlot);
		else
			sum = sum + 40;
		end
		
	end

	sum = sum + TerrainBuilder.GetRandomNumber(10, "River Rand");

	return sum;
end

------------------------------------------------------------------------------
function AddRivers()

	local riverSourceRangeDefault = GlobalParameters.RIVER_SOURCE_RANGE_DEFAULT or 4;
	local seaWaterRangeDefault = GlobalParameters.RIVER_SEA_WATER_RANGE_DEFAULT or 2;
	local plotsPerRiverEdge = GlobalParameters.RIVER_PLOTS_PER_RIVER_EDGE_DEFAULT or 12;
	
	print("Adding rivers ...");
	
	local passConditions = {
		function(plot)
			return (plot:IsHills() or plot:IsMountain());
		end,
		
		function(plot)
			return (not plot:IsCoastalLand()) and (TerrainBuilder.GetRandomNumber(8, "MapGenerator AddRivers") == 0);
		end,
		
		function(plot)
			local area = plot:GetArea();
			return (plot:IsHills() or plot:IsMountain()) and (area:GetRiverEdgeCount() <	((area:GetPlotCount() / plotsPerRiverEdge) + 1));
		end,
		
		function(plot)
			local area = plot:GetArea();
			return (area:GetRiverEdgeCount() < (area:GetPlotCount() / plotsPerRiverEdge) + 1);
		end
	}
	
	for iPass, passCondition in ipairs(passConditions) do
					
		if (iPass <= 2) then
			riverSourceRange = riverSourceRangeDefault;
			seaWaterRange = seaWaterRangeDefault;
		else
			riverSourceRange = (riverSourceRangeDefault / 2);
			seaWaterRange = (seaWaterRangeDefault / 2);
		end
			
		local iW, iH = Map.GetGridSize();

		for i = 0, (iW * iH) - 1, 1 do
			plot = Map.GetPlotByIndex(i);
			local current_x = plot:GetX()
			local current_y = plot:GetY()
			if current_x < 1 or current_x >= g_iW - 2 or current_y < 2 or current_y >= g_iH - 1 or current_x >= g_iW - 1 then
				-- Plot too close to edge, ignore it.
			elseif(not plot:IsWater()) then
				if(passCondition(plot) and plot:IsNaturalWonder() == false and AdjacentToNaturalWonder(plot) == false) then
					if (not Map.FindWater(plot, riverSourceRange, true)) then
						if (not Map.FindWater(plot, seaWaterRange, false)) then
							local inlandCorner = TerrainBuilder.GetInlandCorner(plot);
							if(inlandCorner and plot:IsNaturalWonder() == false and AdjacentToNaturalWonder(plot) == false) then
								local start_x = inlandCorner:GetX()
								local start_y = inlandCorner:GetY()
								local orig_direction;
								if start_y < g_iH / 2 then -- South half of map
									if start_x < g_iW / 3 then -- SW Corner
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHEAST;
									elseif start_x > g_iW * 0.66 then -- SE Corner
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHWEST;
									else -- South, middle
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_NORTH;
									end
								else -- North half of map
									if start_x < g_iW / 3 then -- NW corner
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST;
									elseif start_x > g_iW * 0.66 then -- NE corner
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST;
									else -- North, middle
										orig_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTH;
									end
								end
								
								DoRiver(inlandCorner, nil, orig_direction, nil);
							end
						end
					end
				end			
			end
		end
	end		
end

-------------------------------------------------------------------------------
function AddFeatures()
	print("Adding features ...");

	local args = {};

	-- Get Rainfall setting input by user.
	args.rainfall = MapConfiguration.GetValue(rainfall);
	if args.rainfall == 4 then
		args.rainfall = 1 + TerrainBuilder.GetRandomNumber(3, "Random Rainfall - Lua");
	end
	
	-- Adjust the equator to the vertical middle of the meteor crater
	args.iEquatorAdjustment = math.floor(g_craterCenterY - g_iH / 2);

	-- Reduce features by amount of icy plots that cannot be filled anyways.
	local reductionFactor = (g_iW * g_iH - g_count_icy_plots) / (g_iW * g_iH);
	print("Reducing features from " .. g_iW * g_iH .. " to " .. (g_iW * g_iH - g_count_icy_plots) .. " plots. reductionFactor=", reductionFactor);

	args.iJunglePercent = 12 * reductionFactor;
	args.iForestPercent = 18 * reductionFactor;
	args.iMarshPercent = 3 * reductionFactor;
	args.iOasisPercent = 1 * reductionFactor;

	print("iEquatorAdjustment=", args.iEquatorAdjustment);
	print("iJunglePercent=", args.iJunglePercent);
	print("iForestPercent=", args.iForestPercent);
	print("iMarshPercent=", args.iMarshPercent);
	print("iOasisPercent=", args.iOasisPercent);

	local featuregen = FeatureGenerator.Create(args);

	featuregen:AddFeatures();
end

------------------------------------------------------------------------------
function FeatureGenerator:AddIceAtPlot(plot, iX, iY, lat)
	return
end

------------------------------------------------------------------------------
-- Copied from ResourceGenerator.lua to reduce resources
function ResourceGenerator:__InitResourceData()

	self.iResourcesInDB = 0;
	self.iLuxuriesThisSizeMap = GameInfo.Maps[Map.GetMapSize()].DefaultPlayers * 2;

	-- Get resource value setting input by user.
	if self.resources == 1 then
			self.resources = -3;
	elseif self.resources == 3 then
			self.resources = 3;	
	elseif self.resources == 4 then
		self.resources = TerrainBuilder.GetRandomNumber(9, "Random Resources - Lua") - 4;
	else 
		self.resources = 0;
	end

	self.iTargetPercentage = self.iTargetPercentage + self.resources;

	-- Meteor Survival: Reduce resources by amount of icy plots that have very limited possibilities to be filled
	local reductionFactor = (g_iW * g_iH - g_count_icy_plots) / (g_iW * g_iH);
	print("Reducing resources from " .. g_iW * g_iH .. " to " .. (g_iW * g_iH - g_count_icy_plots) .. " plots. reductionFactor=", reductionFactor);
	self.iTargetPercentage = self.iTargetPercentage * reductionFactor;
	-- --------------------------
    
	for row in GameInfo.Resources() do
		self.eResourceType[self.iResourcesInDB] = row.Hash;
		self.aIndex[self.iResourcesInDB] = row.Index;
		self.eResourceClassType[self.iResourcesInDB] = row.ResourceClassType;
		self.aaPossibleLocs[self.iResourcesInDB] = {};
		self.aaPossibleWaterLocs[self.iResourcesInDB] = {};
		self.aaPossibleLuxLocs[self.iResourcesInDB] = {};
		self.aaPossibleStratLocs[self.iResourcesInDB] = {};
		self.iFrequency[self.iResourcesInDB] = row.Frequency;
		self.iSeaFrequency[self.iResourcesInDB] = row.SeaFrequency;
		self.aPeakEra[self.iResourcesInDB] = row.PeakEra;
	    self.iResourcesInDB = self.iResourcesInDB + 1;
	end
end

------------------------------------------------------------------------------
-- Bugfix for Civ 1.0.0.38 - 1.0.0.110:
-- Even when wonders are excluded with args.Invalid, the wonders with CUSTOM placing logic are still added.
-- Additionally, regardless of invalid or not, the placement is checked in NaturalWonderGenerator.lua,
-- which crashes the script when there is no valid place. Therefore, we replace the function with one that
-- ignores runtime errors.
function CustomCanHaveFeature(pPlot, eFeatureType)
	if unexpected_condition then print("IGNORED ERROR: Could not place wonder!") end
	local aPlots = {};
    local ok, value = pcall(CustomGetMultiTileFeaturePlotList, pPlot, eFeatureType, aPlots);
    if ok then
        return value;
    else
        return false;
    end
end